# Author: Pearu Peterson
# Created: February 2019

import re
from llvmlite import ir
import llvmlite.binding as llvm
from .targetinfo import TargetInfo
from .npy_mathimpl import *  # noqa: F403, F401
from .utils import get_version
if get_version('numba') >= (0, 49):
    from numba.core import codegen, cpu, compiler_lock, \
        registry, typing, compiler, sigutils, cgutils, \
        extending
    from numba.core import types as nb_types
else:
    from numba.targets import codegen, cpu, registry
    from numba import compiler_lock, typing, compiler, \
        sigutils, cgutils, extending
    from numba import types as nb_types

int32_t = ir.IntType(32)


def _lf(lst):
    return lst + [e + 'f' for e in lst] + [e + 'l' for e in lst]


def _f(lst):
    return lst + [e + 'f' for e in lst]


exp_funcs = ['exp', 'exp2', 'expm1', 'log', 'log2', 'log10',
             'log1p', 'ilogb', 'logb']
power_funcs = ['sqrt', 'cbrt', 'hypot', 'pow']
trigonometric_funcs = ['sin', 'cos', 'tan', 'asin', 'acos', 'atan', 'atan2']
hyperbolic_funcs = ['sinh', 'cosh', 'tanh', 'asinh', 'acosh', 'atanh']
nearest_funcs = ['ceil', 'floor', 'trunc', 'round', 'lround', 'llround',
                 'nearbyint', 'rint', 'lrint', 'llrint']
fp_funcs = ['frexp', 'ldexp', 'modf', 'scalbn', 'scalbln', 'nextafter',
            'nexttoward']
classification_funcs = ['fpclassify', 'isfinite', 'isinf', 'isnan',
                        'isnormal', 'signbit']

pymath_funcs = _f(['erf', 'erfc', 'lgamma', 'tgamma'])

fp_funcs = _lf([*exp_funcs, *power_funcs, *trigonometric_funcs,
                *hyperbolic_funcs, *nearest_funcs, *fp_funcs])

libm_funcs = [*fp_funcs, *classification_funcs]

stdio_funcs = ['printf', 'puts', 'fflush']

stdlib_funcs = ['free', 'calloc']


def get_function_dependencies(module, funcname, _deps=None):
    if _deps is None:
        _deps = dict()
    func = module.get_function(funcname)
    assert func.name == funcname, (func.name, funcname)
    for block in func.blocks:
        for instruction in block.instructions:
            if instruction.opcode == 'call':
                name = list(instruction.operands)[-1].name
                f = module.get_function(name)
                if name.startswith('llvm.'):
                    _deps[name] = 'intrinsic'
                elif f.is_declaration:
                    if name in libm_funcs:
                        _deps[name] = 'libm'
                    elif name in stdio_funcs:
                        _deps[name] = 'stdio'
                    elif name in stdlib_funcs:
                        _deps[name] = 'stdlib'
                    elif name in ['allocate_varlen_buffer']:
                        _deps[name] = 'omnisci_internal'
                    elif name in pymath_funcs:
                        _deps[name] = 'pymath'
                    else:
                        _deps[name] = 'undefined'
                else:
                    if name not in _deps:
                        _deps[name] = 'defined'
                        get_function_dependencies(module, name, _deps=_deps)
    return _deps


class JITRemoteCPUCodegen(codegen.JITCPUCodegen):
    # TODO: introduce JITRemoteCodeLibrary?
    _library_class = codegen.JITCodeLibrary

    def __init__(self, name, target_info):
        self.target_info = target_info
        super(JITRemoteCPUCodegen, self).__init__(name)

    def _get_host_cpu_name(self):
        return self.target_info.device_name

    def _get_host_cpu_features(self):
        features = self.target_info.device_features
        if llvm.llvm_version_info[0] < 9:
            # See https://github.com/xnd-project/rbc/issues/45
            for f in ['cx8', 'enqcmd', 'avx512bf16']:
                features = features.replace('+' + f, '').replace('-' + f, '')
        return features

    def _customize_tm_options(self, options):
        super(JITRemoteCPUCodegen, self)._customize_tm_options(options)
        # fix reloc_model as the base method sets it using local target
        if self.target_info.arch.startswith('x86'):
            reloc_model = 'static'
        else:
            reloc_model = 'default'
        options['reloc'] = reloc_model


class RemoteCPUContext(cpu.CPUContext):

    def __init__(self, typing_context, target_info):
        self.target_info = target_info
        super(RemoteCPUContext, self).__init__(typing_context)

    @compiler_lock.global_compiler_lock
    def init(self):
        self.address_size = self.target_info.bits
        self.is32bit = (self.address_size == 32)
        self._internal_codegen = JITRemoteCPUCodegen("numba.exec",
                                                     self.target_info)

        # Map external C functions.
        # nb.core.externals.c_math_functions.install(self)
        # TODO, seems problematic only for 32bit cases

        # Initialize NRT runtime
        # nb.core.argets.cpu.rtsys.initialize(self)
        # TODO: is this needed for LLVM IR generation?

        # import numba.unicode  # unicode support is not relevant here

    # TODO: overwrite load_additional_registries, call_conv?, post_lowering


def make_wrapper(fname, atypes, rtype, cres):
    """Make wrapper function to numba compile result.

    The compilation result contains a function with prototype::

      <status> <function name>(<rtype>** result, <arguments>)

    The make_wrapper adds a wrapper function to compilation result
    library with the following prototype::

      <rtype> <fname>(<arguments>)

    or, if <rtype> is Omnisci Array, then

      void <fname>(<rtype>* <arguments>)

    Parameters
    ----------
    fname : str
      Function name.
    atypes : tuple
      A tuple of argument Numba types.
    rtype : numba.Type
      The return Numba type
    cres : CompileResult
      Numba compilation result.

    """
    fndesc = cres.fndesc
    module = cres.library.create_ir_module(fndesc.unique_name)
    context = cres.target_context
    ll_argtypes = [context.get_value_type(ty) for ty in atypes]
    ll_return_type = context.get_value_type(rtype)

    # TODO: design a API for custom wrapping
    if type(rtype).__name__ == 'ArrayPointer':
        wrapty = ir.FunctionType(ir.VoidType(),
                                 [ll_return_type] + ll_argtypes)
        wrapfn = module.add_function(wrapty, fname)
        builder = ir.IRBuilder(wrapfn.append_basic_block('entry'))
        fnty = context.call_conv.get_function_type(rtype, atypes)
        fn = builder.module.add_function(fnty, cres.fndesc.llvm_func_name)
        status, out = context.call_conv.call_function(
            builder, fn, rtype, atypes, wrapfn.args[1:])
        with cgutils.if_unlikely(builder, status.is_error):
            cgutils.printf(builder,
                           f"rbc: {fname} failed with status code %i\n",
                           status.code)
            builder.ret_void()
        builder.store(builder.load(out), wrapfn.args[0])
        builder.ret_void()
    else:
        wrapty = ir.FunctionType(ll_return_type, ll_argtypes)
        wrapfn = module.add_function(wrapty, fname)
        builder = ir.IRBuilder(wrapfn.append_basic_block('entry'))
        fnty = context.call_conv.get_function_type(rtype, atypes)
        fn = builder.module.add_function(fnty, cres.fndesc.llvm_func_name)
        status, out = context.call_conv.call_function(
            builder, fn, rtype, atypes, wrapfn.args)
        with cgutils.if_unlikely(builder, status.is_error):
            cgutils.printf(builder,
                           f"rbc: {fname} failed with status code %i\n",
                           status.code)
        builder.ret(out)

    cres.library.add_ir_module(module)


def compile_to_LLVM(functions_and_signatures, target: TargetInfo,
                    pipeline_class=compiler.Compiler, debug=False):
    """Compile functions with given signatures to target specific LLVM IR.

    Parameters
    ----------
    functions_and_signatures : list
      Specify a list of Python function and its signatures pairs.
    target : TargetInfo
      Specify target device information.
    debug : bool

    Returns
    -------
    module : llvmlite.binding.ModuleRef
      LLVM module instance. To get the IR string, use `str(module)`.

    """
    target_desc = registry.cpu_target
    if target is None:
        target = TargetInfo.host()
        typing_context = target_desc.typing_context
        target_context = target_desc.target_context
    else:
        typing_context = typing.Context()
        target_context = RemoteCPUContext(typing_context, target)
        # Bring over Array overloads (a hack):
        target_context._defns = target_desc.target_context._defns
        # Fixes rbc issue 74:
        target_desc.typing_context.target_info = target
        target_desc.target_context.target_info = target

    typing_context.target_info = target
    target_context.target_info = target

    codegen = target_context.codegen()
    main_library = codegen.create_library('rbc.irtools.compile_to_IR')
    main_module = main_library._final_module

    flags = compiler.Flags()
    flags.set('no_compile')
    flags.set('no_cpython_wrapper')
    if get_version('numba') >= (0, 49):
        flags.set('no_cfunc_wrapper')

    function_names = []
    for func, signatures in functions_and_signatures:
        for sig in signatures:
            fname = func.__name__ + sig.mangling()
            function_names.append(fname)
            args, return_type = sigutils.normalize_signature(
                sig.tonumba(bool_is_int8=True))
            cres = compiler.compile_extra(typingctx=typing_context,
                                          targetctx=target_context,
                                          func=func,
                                          args=args,
                                          return_type=return_type,
                                          flags=flags,
                                          library=main_library,
                                          locals={},
                                          pipeline_class=pipeline_class)
            make_wrapper(fname, args, return_type, cres)

    seen = set()
    for _library in main_library._linking_libraries:
        if _library not in seen:
            seen.add(_library)
            main_module.link_in(
                _library._get_module_for_linking(), preserve=True,
            )

    main_library._optimize_final_module()

    # Catch undefined functions:
    used_functions = set(function_names)
    for fname in function_names:
        deps = get_function_dependencies(main_module, fname)
        for fn, descr in deps.items():
            used_functions.add(fn)
            if descr == 'undefined':
                if fn.startswith('numba_') and target.has_numba:
                    continue
                if fn.startswith('Py') and target.has_cpython:
                    continue
                raise RuntimeError('function `%s` is undefined' % (fn))

    # for global_variable in main_module.global_variables:
    #    global_variable.linkage = llvm.Linkage.private

    unused_functions = [f.name for f in main_module.functions
                        if f.name not in used_functions]

    if debug:
        print('compile_to_IR: the following functions are used')
        for fname in used_functions:
            lf = main_module.get_function(fname)
            print('  [ALIVE]', fname, 'with', lf.linkage)

    if unused_functions:
        if debug:
            print('compile_to_IR: the following functions are not used'
                  ' and will be removed:')
        for fname in unused_functions:
            lf = main_module.get_function(fname)
            if lf.is_declaration:
                # if the function is a declaration,
                # we just put the linkage as external
                lf.linkage = llvm.Linkage.external
            else:
                # but if the function is not a declaration,
                # we change the linkage to private
                lf.linkage = llvm.Linkage.private
            if debug:
                print('  [DEAD]', fname, 'with', lf.linkage)

        main_library._optimize_final_module()
    # TODO: determine unused global_variables and struct_types

    main_module.verify()
    main_library._finalized = True

    main_module.triple = target.triple
    main_module.data_layout = target.datalayout

    return main_module


def compile_IR(ir):
    """Return execution engine with IR compiled in.

    Parameters
    ----------
    ir : str
      Specify LLVM IR code as a string.

    Returns
    -------
    engine :
      Execution engine.


    Examples
    --------

        To get the address of the compiled functions, use::

          addr = engine.get_function_address("<function name>")
    """
    triple = re.search(
        r'target\s+triple\s*=\s*"(?P<triple>[-\d\w\W_]+)"\s*$',
        ir, re.M).group('triple')

    # Create execution engine
    target = llvm.Target.from_triple(triple)
    target_machine = target.create_target_machine()
    backing_mod = llvm.parse_assembly("")
    engine = llvm.create_mcjit_compiler(backing_mod, target_machine)

    # Create LLVM module and compile
    mod = llvm.parse_assembly(ir)
    mod.verify()
    engine.add_module(mod)
    engine.finalize_object()
    engine.run_static_constructors()

    return engine


def cg_fflush(builder):
    int8_t = ir.IntType(8)
    fflush_fnty = ir.FunctionType(int32_t, [int8_t.as_pointer()])
    fflush_fn = builder.module.get_or_insert_function(
        fflush_fnty, name='fflush')

    builder.call(fflush_fn, [int8_t.as_pointer()(None)])


@extending.intrinsic
def fflush(typingctx):
    """fflush that can be called from Numba jit-decorated functions.
    """
    sig = nb_types.void(nb_types.void)

    def codegen(context, builder, signature, args):
        cg_fflush(builder)

    return sig, codegen


@extending.intrinsic
def printf(typingctx, format_type, *args):
    """printf that can be called from Numba jit-decorated functions.
    """
    if isinstance(format_type, nb_types.StringLiteral):
        sig = nb_types.void(format_type, nb_types.BaseTuple.from_types(args))

        def codegen(context, builder, signature, args):
            cgutils.printf(builder, format_type.literal_value, *args[1:])
            cg_fflush(builder)

        return sig, codegen



"""Implement Omnisci Column type support

Omnisci Column type is the type of input/output column arguments in
UDTFs.
"""

__all__ = ['ColumnPointer', 'OutputColumn', 'Column',
           'OmnisciOutputColumnType', 'OmnisciColumnType',
           'output_column_type_converter', 'column_type_converter']

from rbc.utils import get_version
from .omnisci_buffer import (
    BufferPointer, Buffer, BufferPointerModel,
    buffer_type_converter, OmnisciBufferType,
)

if get_version('numba') >= (0, 49):
    from numba.core import datamodel
else:
    from numba import datamodel


class OmnisciColumnType(OmnisciBufferType):
    """Omnisci Column type for RBC typesystem.
    """


class OmnisciOutputColumnType(OmnisciBufferType):
    """Omnisci OutputColumn type for RBC typesystem.

    OutputColumn is the same as Column but introduced to distinguish
    the input and output arguments of UDTFs.
    """


class ColumnPointer(BufferPointer):
    """Type class for pointers to :code:`Omnisci Column<T>` structure."""


class Column(Buffer):
    pass


class OutputColumn(Column):
    pass


@datamodel.register_default(ColumnPointer)
class ColumnPointerModel(BufferPointerModel):
    pass


def column_type_converter(target_info, obj):
    """Return Type instance corresponding to Omnisci :code:`Column<T>` type.

    :code:`Omnisci Column<T>` is defined as follows (using C++ syntax)::

      template<typename T>
      struct Column {
        T* ptr;
        size_t sz;
      }

    See :code:`buffer_type_converter` for details.
    """
    return buffer_type_converter(
        target_info, obj, OmnisciColumnType, 'Column', ColumnPointer,
        extra_members=[])


def output_column_type_converter(target_info, obj):
    """Return Type instance corresponding to Omnisci :code:`OutputColumn<T>` type.

    See :code:`column_type_converter` for implementation detail.
    """
    return buffer_type_converter(
        target_info, obj, OmnisciOutputColumnType, 'OutputColumn',
        ColumnPointer, extra_members=[])



import os
from collections import defaultdict
import pytest


rbc_omnisci = pytest.importorskip('rbc.omniscidb')
available_version, reason = rbc_omnisci.is_available()
if available_version and available_version < (5, 4):
    reason = ('New-style UDTFs (with Column arguments) are available'
              ' for omniscidb 5.4 or newer, '
              'currently connected to omniscidb '
              + '.'.join(map(str, available_version)))
    available_version = ()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def omnisci():
    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)
    table_name = os.path.splitext(os.path.basename(__file__))[0]
    m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))

    sqltypes = ['FLOAT', 'DOUBLE', 'TINYINT', 'SMALLINT', 'INT', 'BIGINT',
                'BOOLEAN', 'DOUBLE']
    colnames = ['f4', 'f8', 'i1', 'i2', 'i4', 'i8', 'b', 'd']
    table_defn = ',\n'.join('%s %s' % (n, t)
                            for t, n in zip(sqltypes, colnames))
    m.sql_execute(
        'CREATE TABLE IF NOT EXISTS {table_name} ({table_defn});'
        .format(**locals()))

    data = defaultdict(list)
    for i in range(5):
        for j, n in enumerate(colnames):
            if n == 'b':
                data[n].append(i % 2 == 0)
            elif n == 'd':
                data[n].append(i + 1.5)
            else:
                data[n].append(i)

    m.load_table_columnar(table_name, **data)
    m.table_name = table_name
    yield m
    m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))


def test_sizer_row_multiplier_orig(omnisci):
    if omnisci.has_cuda:
        pytest.skip('crashes CUDA enabled omniscidb server'
                    ' [rbc issue 147]')
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(Column<double>, int64|sizer=RowMultiplier,'
             ' OutputColumn<double>)')
    def my_row_copier_mul(x, m, y):
        input_row_count = len(x)
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                y[j] = x[i] * 2
        return m * input_row_count

    descr, result = omnisci.sql_execute(
        'select * from table(my_row_copier_mul(cursor(select f8 '
        'from {omnisci.table_name}), 2));'
        .format(**locals()))

    for i, r in enumerate(result):
        assert r == ((i % 5) * 2,)


def test_sizer_row_multiplier_param1(omnisci):
    if omnisci.has_cuda:
        pytest.skip('crashes CUDA enabled omniscidb server'
                    ' [rbc issue 147]')

    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(Column<double>, double, int32, int64|sizer=RowMultiplier,'
             'OutputColumn<double>)')
    def my_row_copier_mul_param1(x, alpha, beta, m, y):
        input_row_count = len(x)
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                y[j] = x[i] * alpha + beta
        return m * input_row_count

    alpha = 3

    descr, result = omnisci.sql_execute(
        'select * from table(my_row_copier_mul_param1('
        'cursor(select f8 from {omnisci.table_name}),'
        'cast({alpha} as double),'
        'cast(4 as int), 2));'
        .format(**locals()))

    for i, r in enumerate(result):
        assert r == ((i % 5) * alpha + 4,)


def test_sizer_row_multiplier_param2(omnisci):
    if omnisci.has_cuda:
        pytest.skip('crashes CUDA enabled omniscidb server'
                    ' [rbc issue 147]')

    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(double, Column<double>, int32, int64|sizer=RowMultiplier,'
             'OutputColumn<double>)')
    def my_row_copier_mul_param2(alpha, x, beta, m, y):
        input_row_count = len(x)
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                y[j] = x[i] * alpha + beta
        return m * input_row_count

    alpha = 3

    descr, result = omnisci.sql_execute(
        'select * from table(my_row_copier_mul_param2('
        'cast({alpha} as double),'
        'cursor(select f8 from {omnisci.table_name}),'
        'cast(4 as int), 2));'
        .format(**locals()))

    for i, r in enumerate(result):
        assert r == ((i % 5) * alpha + 4,)


@pytest.mark.skipif(
    True,
    reason=(
        "test requires omniscidb with constant parameter"
        " support (got %s) [issue 124]" % (
            available_version,)))
def test_sizer_constant_parameter(omnisci):
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(Column<double>, int64|sizer=ConstantParameter,'
             ' Column<double>)')
    def my_row_copier_cp(x, m, y):
        input_row_count = len(x)
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                y[j] = x[i] * 2
        return m

    descr, result = omnisci.sql_execute(
        'select * from table(my_row_copier_cp(cursor(select f8 '
        'from {omnisci.table_name}), 5));'
        .format(**locals()))

    for i, r in enumerate(result):
        assert r == ((i % 5) * 2,)


@pytest.mark.skipif(
    available_version < (5, 4),
    reason=(
        "test requires omniscidb v 5.4 or newer (got %s) [issue 148]" % (
            available_version,)))
def test_rowmul_add_columns(omnisci):
    if omnisci.has_cuda:
        pytest.skip('crashes CUDA enabled omniscidb server'
                    ' [issue 147]')

    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(Column<double>, Column<double>, double,'
             ' int64|sizer=RowMultiplier, OutputColumn<double>)')
    def add_columns(x, y, alpha, m, r):
        input_row_count = len(x)
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                r[j] = x[i] + alpha * float(y[i])
        return m * input_row_count

    alpha = 2.5

    descr, result = omnisci.sql_execute(
        'select * from table(add_columns('
        'cursor(select f8 from {omnisci.table_name}),'
        ' cursor(select d from {omnisci.table_name}),'
        ' cast({alpha} as double), 1));'
        .format(**locals()))

    for i, r in enumerate(result):
        assert r == (i + alpha * (i + 1.5),)


@pytest.mark.skipif(
    available_version < (5, 4),
    reason=(
        "test requires omniscidb with multiple output"
        " columns support (got %s) [issue 150]" % (
            available_version,)))
def test_rowmul_return_two_columns(omnisci):
    if omnisci.has_cuda:
        pytest.skip('crashes CUDA enabled omniscidb server'
                    ' [rbc issue 147]')
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(Column<double>, int64|sizer=RowMultiplier,'
             ' OutputColumn<double>, OutputColumn<double>)')
    def ret_columns(x, m, y, z):
        input_row_count = len(x)
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                y[j] = 2 * x[i]
                z[j] = 3 * x[i]
        return m * input_row_count

    descr, result = omnisci.sql_execute(
        'select * from table(ret_columns('
        f'cursor(select f8 from {omnisci.table_name}), 1));')

    for i, r in enumerate(result):
        assert r[0] == 2 * i
        assert r[1] == 3 * i

    descr, result = omnisci.sql_execute(
        'select out1, out0 from table(ret_columns('
        f'cursor(select f8 from {omnisci.table_name}), 1));')

    for i, r in enumerate(result):
        assert r[1] == 2 * i
        assert r[0] == 3 * i



"""Implement Omnisci Array type support
"""

__all__ = ['ArrayPointer', 'Array', 'omnisci_array_constructor',
           'array_type_converter']

import warnings
from collections import defaultdict
from llvmlite import ir
from rbc import typesystem
from rbc.utils import get_version
from .omnisci_buffer import (BufferPointer, Buffer, BufferPointerModel,
                             buffer_type_converter, OmnisciBufferType)

if get_version('numba') >= (0, 49):
    from numba.core import datamodel, cgutils, extending, types
else:
    from numba import datamodel, cgutils, extending, types


int8_t = ir.IntType(8)
int32_t = ir.IntType(32)
int64_t = ir.IntType(64)
void_t = ir.VoidType()


class OmnisciArrayType(OmnisciBufferType):
    """Omnisci Array type for RBC typesystem.
    """


class ArrayPointer(BufferPointer):
    """Type class for pointers to :code:`Omnisci Array<T>` structure."""


class Array(Buffer):
    pass


@datamodel.register_default(ArrayPointer)
class ArrayPointerModel(BufferPointerModel):
    pass


builder_buffers = defaultdict(list)


@extending.lower_builtin(Array, types.Integer, types.StringLiteral)
@extending.lower_builtin(Array, types.Integer, types.NumberClass)
def omnisci_array_constructor(context, builder, sig, args):
    if not context.target_info.is_cpu:
        warnings.warn(
            f'allocating arrays in {context.target_info.name}'
            ' is not supported')
    ptr_type, sz_type, null_type = sig.return_type.dtype.members

    # zero-extend the element count to int64_t
    assert isinstance(args[0].type, ir.IntType), (args[0].type)
    element_count = builder.zext(args[0], int64_t)
    element_size = int64_t(ptr_type.dtype.bitwidth // 8)

    '''
    QueryEngine/ArrayOps.cpp:
    int8_t* allocate_varlen_buffer(int64_t element_count, int64_t element_size)
    '''
    alloc_fnty = ir.FunctionType(int8_t.as_pointer(), [int64_t, int64_t])
    # see https://github.com/xnd-project/rbc/issues/75
    alloc_fn = builder.module.get_or_insert_function(
        alloc_fnty, name="calloc")
    ptr8 = builder.call(alloc_fn, [element_count, element_size])
    builder_buffers[builder].append(ptr8)
    ptr = builder.bitcast(ptr8, context.get_value_type(ptr_type))
    is_null = context.get_value_type(null_type)(0)

    # construct array
    fa = cgutils.create_struct_proxy(sig.return_type.dtype)(context, builder)
    fa.ptr = ptr              # T*
    fa.sz = element_count     # size_t
    fa.is_null = is_null      # int8_t
    return fa._getpointer()


@extending.type_callable(Array)
def type_omnisci_array(context):
    def typer(size, dtype):
        if isinstance(dtype, types.StringLiteral):
            typ = 'Array<{}>'.format(dtype.literal_value)
        elif isinstance(dtype, types.NumberClass):
            typ = 'Array<{}>'.format(dtype.dtype)
        else:
            raise NotImplementedError(repr(dtype))
        atyp = array_type_converter(context.target_info, typ)
        if atyp is not None:
            return atyp.tonumba()
        raise NotImplementedError((dtype, typ))
    return typer


@extending.intrinsic
def omnisci_array_is_null_(typingctx, data):
    sig = types.int8(data)

    def codegen(context, builder, signature, args):

        rawptr = cgutils.alloca_once_value(builder, value=args[0])
        ptr = builder.load(rawptr)

        return builder.load(builder.gep(ptr, [int32_t(0), int32_t(2)]))

    return sig, codegen


@extending.overload_method(ArrayPointer, 'is_null')
def omnisci_array_is_null(x):
    if isinstance(x, ArrayPointer):
        def impl(x):
            return omnisci_array_is_null_(x)
        return impl


def array_type_converter(target_info, obj):
    """Return Type instance corresponding to :code:`Omnisci Array<T>` type.

    :code:`Omnisci Array<T>` is defined as follows (using C++ syntax)::

      template<typename T>
      struct Array {
        T* ptr;
        size_t sz;
        bool is_null;
      }

    See :code:`buffer_type_converter` for details.
    """
    buffer_type = buffer_type_converter(
        target_info, obj, OmnisciArrayType,
        'Array', ArrayPointer,
        extra_members=[
            typesystem.Type.fromstring('bool is_null',
                                       target_info=target_info)])
    if buffer_type is not None:
        return buffer_type.pointer()
    return buffer_type



"""Implement Buffer type as a base class to Omnisci Array and Column types.

Omnisci Buffer represents the following structure:

  template<typename T>
  struct Buffer {
    T* ptr;
    size_t sz;
    ...
  }

that is, a structure that has at least two members where the first is
a pointer to some data type and the second is the size of the buffer.

This module implements the support for the following Python operators:

  len
  __getitem__
  __setitem__

to provide a minimal functionality for accessing and manipulating the
Omnisci buffer objects from UDF/UDTFs.
"""


import operator
from llvmlite import ir
from rbc import typesystem
from rbc.utils import get_version
if get_version('numba') >= (0, 49):
    from numba.core import datamodel, cgutils, extending, types
else:
    from numba import datamodel, cgutils, extending, types


int8_t = ir.IntType(8)
int32_t = ir.IntType(32)
int64_t = ir.IntType(64)
void_t = ir.VoidType()


class OmnisciBufferType(typesystem.Type):
    """Typesystem type class for Omnisci buffer structures.
    """

    @property
    def _buffer_typename(self):
        return self._params['typename']

    def pointer(self):
        ptr_type = self._params.get('pointer')
        if ptr_type is not None:
            return ptr_type
        return typesystem.Type.pointer(self)


class BufferType(types.Type):
    """Numba type class for Omnisci buffer structures.
    """


class BufferPointer(types.Type):
    """Numba type class for pointers to Omnisci buffer structures.

    We are not deriving from CPointer because BufferPointer getitem is
    used to access the data stored in Buffer ptr member.
    """
    mutable = True

    def __init__(self, dtype, eltype):
        self.dtype = dtype    # struct dtype
        self.eltype = eltype  # buffer element dtype
        name = "(%s)*" % dtype
        types.Type.__init__(self, name)

    @property
    def key(self):
        return self.dtype


class Buffer(object):
    """Represents Omnisci Buffer that can be constructed within UDF/UDTFs.
    """


class BufferPointerModel(datamodel.models.PointerModel):
    """Base class for Omnisci buffer pointer models.

    Subclasses should register the model for the corresponding
    BufferPointer subclass, for instance::

      @datamodel.register_default(ArrayPointer)
      class ArrayPointerModel(BufferPointerModel):
          pass
    """


@extending.intrinsic
def omnisci_buffer_ptr_len_(typingctx, data):
    sig = types.int64(data)

    def codegen(context, builder, signature, args):
        data, = args
        rawptr = cgutils.alloca_once_value(builder, value=data)
        struct = builder.load(builder.gep(rawptr,
                                          [int32_t(0)]))
        return builder.load(builder.gep(
            struct, [int32_t(0), int32_t(1)]))
    return sig, codegen


@extending.intrinsic
def omnisci_buffer_len_(typingctx, data):
    sig = types.int64(data)

    def codegen(context, builder, signature, args):
        data, = args
        assert data.opname == 'load'
        struct = data.operands[0]
        return builder.load(builder.gep(
            struct, [int32_t(0), int32_t(1)]))

    return sig, codegen


@extending.overload(len)
def omnisci_buffer_len(x):
    if isinstance(x, BufferPointer):
        return lambda x: omnisci_buffer_ptr_len_(x)
    if isinstance(x, BufferType):
        return lambda x: omnisci_buffer_len_(x)


@extending.intrinsic
def omnisci_buffer_ptr_getitem_(typingctx, data, index):
    sig = data.eltype(data, index)

    def codegen(context, builder, signature, args):
        data, index = args
        rawptr = cgutils.alloca_once_value(builder, value=data)
        buf = builder.load(builder.gep(rawptr, [int32_t(0)]))
        ptr = builder.load(builder.gep(
            buf, [int32_t(0), int32_t(0)]))
        res = builder.load(builder.gep(ptr, [index]))

        return res
    return sig, codegen


@extending.intrinsic
def omnisci_buffer_getitem_(typingctx, data, index):
    eltype = data.members[0].dtype
    sig = eltype(data, index)

    def codegen(context, builder, signature, args):
        data, index = args
        assert data.opname == 'load'
        buf = data.operands[0]

        ptr = builder.load(builder.gep(
            buf, [int32_t(0), int32_t(0)]))
        res = builder.load(builder.gep(ptr, [index]))

        return res
    return sig, codegen


@extending.overload(operator.getitem)
def omnisci_buffer_getitem(x, i):
    if isinstance(x, BufferPointer):
        return lambda x, i: omnisci_buffer_ptr_getitem_(x, i)
    if isinstance(x, BufferType):
        return lambda x, i: omnisci_buffer_getitem_(x, i)


@extending.intrinsic
def omnisci_buffer_ptr_setitem_(typingctx, data, index, value):
    sig = types.none(data, index, value)

    def codegen(context, builder, signature, args):
        zero = int32_t(0)

        data, index, value = args

        rawptr = cgutils.alloca_once_value(builder, value=data)
        ptr = builder.load(rawptr)

        buf = builder.load(builder.gep(ptr, [zero, zero]))
        builder.store(value, builder.gep(buf, [index]))

    return sig, codegen


@extending.intrinsic
def omnisci_buffer_setitem_(typingctx, data, index, value):
    sig = types.none(data, index, value)

    def codegen(context, builder, signature, args):
        zero = int32_t(0)

        data, index, value = args

        assert data.opname == 'load'
        buf = data.operands[0]

        ptr = builder.load(builder.gep(
            buf, [zero, zero]))

        builder.store(value, builder.gep(ptr, [index]))

    return sig, codegen


@extending.overload(operator.setitem)
def omnisci_buffer_setitem(a, i, v):
    if isinstance(a, BufferPointer):
        return lambda a, i, v: omnisci_buffer_ptr_setitem_(a, i, v)
    if isinstance(a, BufferType):
        return lambda a, i, v: omnisci_buffer_setitem_(a, i, v)


def make_buffer_ptr_type(buffer_type, buffer_ptr_cls):
    ptr_type = buffer_type.pointer()

    numba_type = buffer_type.tonumba(bool_is_int8=True)
    numba_eltype = buffer_type[0][0].tonumba(bool_is_int8=True)
    numba_type_ptr = buffer_ptr_cls(numba_type, numba_eltype)

    ptr_type._params['tonumba'] = numba_type_ptr
    ptr_type._params['typename'] = buffer_type._params['typename'] + '*'
    return ptr_type


def buffer_type_converter(target_info, obj, typesystem_buffer_type,
                          buffer_typename, buffer_ptr_cls,
                          extra_members=[]):
    """Template function for converting typesystem Type instances to
    Omnisci Buffer Type instance.

    Note
    ----
    It is assumed that the caller of this function is registered to
    TargetInfo. Otherwise, the convertion from a string object will
    not function correctly.

    Parameters
    ----------
    obj : {str, typesystem.Type}
      Custom atomic type corresponding to Omnisci Buffer.
    typesystem_buffer_type : {OmnisciBufferType subclass}
    buffer_typename : str
      The name of buffer type.
    buffer_ptr_cls : {BufferPointer subclass}
    extra_members : {list of typesystem Type instances}
      Additional members of the buffer structure.

    """
    if isinstance(obj, str):
        obj = typesystem.Type(obj,)

    if not isinstance(obj, typesystem.Type):
        raise NotImplementedError(type(obj))

    # A valid obj for buffer type is in the form
    #   typesystem.Type('BufferTypeName<ElementType>', )
    # In all other cases refuse by returning None.

    name, params = obj.get_name_and_parameters()
    if name is None or name != buffer_typename:
        return
    assert len(params) == 1, params
    t = typesystem.Type.fromstring(params[0], target_info)
    if not t.is_concrete:
        return

    assert issubclass(buffer_ptr_cls, BufferPointer)
    assert issubclass(typesystem_buffer_type, OmnisciBufferType)

    # Construct buffer type as a struct with ptr and sz as members.
    ptr_t = typesystem.Type(t, '*', name='ptr')
    size_t = typesystem.Type.fromstring('size_t sz',
                                        target_info=target_info)
    buffer_type = typesystem_buffer_type(
        ptr_t,
        size_t,
        *extra_members
    )

    # Converting buffer type to numba type results BufferType instance
    # that simplifies checking types when overloading generic methods
    # like len, getitem, setitem, etc.
    buffer_type._params['numba.Type'] = BufferType

    # Create alias to buffer type
    typename = '%s<%s>' % (buffer_typename, t.toprototype())
    buffer_type._params['typename'] = typename

    # In omniscidb, boolean values are stored as int8 because
    # boolean has three states: false, true, and null.
    numba_type = buffer_type.tonumba(bool_is_int8=True)

    buffer_type._params['tonumba'] = numba_type
    buffer_type._params['pointer'] = make_buffer_ptr_type(buffer_type,
                                                          buffer_ptr_cls)

    return buffer_type
